'http://www.dataidee.com/ping/ping.htm

Imports System.Net
Imports System.Net.Sockets
Imports System.Threading

'
' A class that uses a ICMP ping to verify if a remote PC is alive
'

Public Class Ping

#If V7 Then
    ' socket and event
    Private pingSocket As Socket
    Private isOpen As Boolean
    Private readComplete As ManualResetEvent
    Private lastSequenceNr As Byte

    ' ping command and result fields
    Private pingCommand As Byte()
    Private pingResult As Byte()

    Public Sub New()

        ' create command field
        ReDim pingCommand(7)
        pingCommand(0) = 8      ' type
        pingCommand(1) = 0      ' subtype
        pingCommand(2) = 0      ' checksum
        pingCommand(3) = 0
        pingCommand(4) = 1      ' identifier
        pingCommand(5) = 0
        pingCommand(6) = 0      ' sequence number
        pingCommand(7) = 0

        ' create result field
        ReDim pingResult(pingCommand.Length + 1000)
    End Sub

    Public Sub Open()

        'create socket and connect
        pingSocket = New Socket(AddressFamily.InterNetwork, SocketType.Raw, ProtocolType.Icmp)
        isOpen = True

        ' create read complete event
        readComplete = New ManualResetEvent(False)
    End Sub

    Public Sub Close()
        ' clear open flag
        isOpen = False

        ' close socket
        pingSocket.Close()

        ' close event
        readComplete.Close()
    End Sub

    Public Function Ping(ByVal RemoteAddress As String) As Integer
        Return Ping(RemoteAddress, 1000)
    End Function

    Public Function Ping(ByVal RemoteAddress As String, ByVal MilliSeconds As Integer) As Integer
        Dim ts_in, ts_out As TimeSpan
        Dim ExplicitOpen As Boolean = False
        Static PreviousAddress As String
        Dim i As Integer

        If isOpen = False Then
            Open()
            ExplicitOpen = True
        End If

        If PreviousAddress <> RemoteAddress Then
            ' start a new sequence number
            lastSequenceNr = 0
            PreviousAddress = RemoteAddress
        End If

        ts_in = New TimeSpan(0, 0, 0, 0, MilliSeconds)
        ' you get 3 tries
        For i = 0 To 2
            ts_out = Send(RemoteAddress, ts_in)
            If ts_out.CompareTo(TimeSpan.MaxValue) < 0 Then
                Exit For
            End If
        Next

        If ExplicitOpen Then
            Close()
        End If

        If ts_out.CompareTo(TimeSpan.MaxValue) = 0 Then
            Return MilliSeconds + 1
        Else
            Return CInt(ts_out.TotalMilliseconds)
        End If

    End Function

    Public Function Send(ByVal address As String, ByVal timeout As TimeSpan) As TimeSpan
        Dim timeSend As DateTime
        Dim iSend As Integer

        ' empty result buffer
        While (pingSocket.Available > 0)
            pingSocket.Receive(pingResult, Math.Min(pingSocket.Available, pingResult.Length), SocketFlags.None)
        End While

        'reset event
        readComplete.Reset()

        ' save start time
        timeSend = DateTime.Now

        ' complete ping command
        pingCommand(6) = lastSequenceNr
        lastSequenceNr += CByte(1)
        SetChecksum(pingCommand)

        ' send ping command, start receive command
        Try
            If address.ToLower.IndexOfAny("abcdefghijklmnopqrstuvwxyz-".ToCharArray) >= 0 Then
                iSend = pingSocket.SendTo(pingCommand, New IPEndPoint(Dns.GetHostByName(address).AddressList(0), 0))
            Else
                iSend = pingSocket.SendTo(pingCommand, New IPEndPoint(IPAddress.Parse(address), 0))
            End If

            If iSend <> pingCommand.Length Then
                Return TimeSpan.MaxValue
            End If
        Catch ex As Exception
            Return TimeSpan.MaxValue
        End Try
        pingSocket.BeginReceive(pingResult, 0, pingResult.Length, SocketFlags.None, New AsyncCallback(AddressOf CallBack), Nothing)

        ' wait until data received or timeout
        If (readComplete.WaitOne(timeout, False)) Then
            ' check result
            If ((pingResult(20) = 0) And _
                (pingCommand(4) = pingResult(24)) And (pingCommand(5) = pingResult(25)) And _
                (pingCommand(6) = pingResult(26)) And (pingCommand(7) = pingResult(27))) Then

                'return time delay
                Return DateTime.Now.Subtract(timeSend)
            End If
        End If

        ' return timeout
        Return TimeSpan.MaxValue
    End Function

    Private Sub CallBack(ByVal result As IAsyncResult)
        If isOpen Then
            Try
                pingSocket.EndReceive(result)
            Catch ex As Exception
            End Try
            Try
                readComplete.Set()
            Catch ex As Exception
            End Try
        End If
    End Sub

    Private Sub SetChecksum(ByRef tel() As Byte)
        Dim i, cs As Integer

        ' reset old checksum
        tel(2) = 0
        tel(3) = 0

        ' calculate new checksum
        For i = 0 To pingCommand.Length - 1 Step 2
            cs += BitConverter.ToInt16(pingCommand, i)
        Next
        cs = Not ((cs And &HFFFF) + (cs \ &H10000))

        ' set new checksum
        tel(2) = BitConverter.GetBytes(cs)(0)
        tel(3) = BitConverter.GetBytes(cs)(1)
    End Sub
#Else

    Public Function Ping(ByVal RemoteAddress As String) As Integer
        Return Ping(RemoteAddress, 1000)
    End Function

    ' Doing a ping operation on .Net 2.0 is a lot easier
    Public Function Ping(ByVal RemoteAddress As String, ByVal MilliSeconds As Integer) As Integer
        Dim i As Integer

        ' we try  up to 3 times
        Try
            For i = 0 To 2
                If My.Computer.Network.Ping(RemoteAddress, MilliSeconds) Then
                    Return 0
                End If
            Next
        Catch ex As Exception
        End Try

        Return MilliSeconds + 1
    End Function
#End If
End Class
