Imports System.Text
Imports System.Runtime.InteropServices
Imports System.ComponentModel
Imports System.Collections.Specialized

'
' A registry encapsulation class similar to Windows.Win32.RegistryKey that
' includes an extra RegWow64Options parameter that allows .Net applications
' to pick which registry "view" they want to see on 64-bit operating systems.
' It also allows .Net 1.1 applications to use the DoNotExpandEnvironmentNames
' feature.

' Note: I was lazy, and only implemented the features required by this
' application (but I left some breadcrumbs behind to remind me of the 
' features that need to be added).

'
' Public Enumerations
'
Public Enum RegistryHive As Integer
    LocalMachine = Microsoft.Win32.RegistryHive.LocalMachine
    ClassesRoot = Microsoft.Win32.RegistryHive.ClassesRoot
    CurrentUser = Microsoft.Win32.RegistryHive.CurrentUser
    Users = Microsoft.Win32.RegistryHive.Users
    CurrentConfig = Microsoft.Win32.RegistryHive.CurrentConfig
End Enum

Public Enum RegWow64Options As Integer
    None = 0
    KEY_WOW64_64KEY = &H100
    KEY_WOW64_32KEY = &H200
End Enum

Public Enum RegistryValueOptions As Integer
    None = 0
    DoNotExpandEnvironmentNames
End Enum

Public Enum RegistryRights As Integer
    ReadKey = 131097
    WriteKey = 131078
End Enum

Public Enum RegistryValueKind As Integer
    Unknown = 0
    [String] = 1
    ExpandString = 2
    Binary = 3
    DWord = 4
    MultiString = 7
    QWord = 11
End Enum

' Future: RegistryKeyPermissionCheck

Public Class RegistryKey

#Region "API Region"
    Const REG_NONE As Integer = 0
    Const REG_SZ As Integer = 1
    Const REG_EXPAND_SZ As Integer = 2
    Const REG_BINARY As Integer = 3
    Const REG_DWORD As Integer = 4
    Const REG_DWORD_BIG_ENDIAN As Integer = 5
    Const REG_LINK As Integer = 6
    Const REG_MULTI_SZ As Integer = 7
    Const REG_RESOURCE_LIST As Integer = 8
    Const REG_FULL_RESOURCE_DESCRIPTOR As Integer = 9
    Const REG_RESOURCE_REQUIREMENTS_LIST As Integer = 10
    Const REG_QWORD As Integer = 11

    Const ERROR_MORE_DATA As Integer = 234
    Const MAX_REG_KEYNAME_SIZE As Integer = 255

    'LONG RegOpenKeyEx(
    '  HKEY hKey,
    '  LPCTSTR lpSubKey,
    '  DWORD ulOptions,
    '  REGSAM samDesired,
    '  PHKEY phkResult
    ');
    Declare Auto Function RegOpenKeyEx Lib "Advapi32" ( _
        ByVal hKey As IntPtr, _
        ByVal lpSubKey As String, _
        ByVal ulOptions As Integer, _
        ByVal samDesired As Integer, _
        ByRef phkResults As IntPtr _
    ) As Integer

    'LONG RegCloseKey(
    '  HKEY hKey
    ');
    Declare Function RegCloseKey Lib "Advapi32" ( _
        ByVal hKey As IntPtr _
    ) As Integer

    'LONG RegQueryValueEx(
    '  HKEY hKey,
    '  LPCTSTR lpValueName,
    '  LPDWORD lpReserved,
    '  LPDWORD lpType,
    '  LPBYTE lpData,
    '  LPDWORD lpcbData
    ');
    Declare Auto Function RegQueryValueEx Lib "Advapi32" ( _
        ByVal hKey As IntPtr, _
        ByVal lpValueName As String, _
        ByVal lpReserved As IntPtr, _
        ByRef lpType As Integer, _
        ByVal lpData As IntPtr, _
        ByRef lpcbData As Integer _
    ) As Integer

    'LONG RegSetValueEx(
    '  HKEY hKey,
    '  LPCTSTR lpValueName,
    '  DWORD Reserved,
    '  DWORD dwType,
    '  const BYTE* lpData,
    '  DWORD cbData
    ');
    Declare Auto Function RegSetValueEx Lib "Advapi32" ( _
        ByVal hKey As IntPtr, _
        ByVal lpValueName As String, _
        ByVal Reserved As Integer, _
        ByVal dwType As Integer, _
        ByVal lpData As IntPtr, _
        ByVal cbData As Integer _
    ) As Integer

    'LONG RegEnumKeyEx(
    '  HKEY hKey,
    '  DWORD dwIndex,
    '  LPTSTR lpName,
    '  LPDWORD lpcName,
    '  LPDWORD lpReserved,
    '  LPTSTR lpClass,
    '  LPDWORD lpcClass,
    '  PFILETIME lpftLastWriteTime
    ');
    Declare Auto Function RegEnumKeyEx Lib "Advapi32" ( _
        ByVal hKey As IntPtr, _
        ByVal dwIndex As Integer, _
        ByVal lpName As StringBuilder, _
        ByRef lpcName As Integer, _
        ByVal lpReserved As IntPtr, _
        ByVal lpClass As IntPtr, _
        ByVal lpcClass As IntPtr, _
        ByVal lpftLastWriteTime As IntPtr _
    ) As Integer

    'LONG RegEnumValue(
    '  HKEY hKey,
    '  DWORD dwIndex,
    '  LPTSTR lpValueName,
    '  LPDWORD lpcValueName,
    '  LPDWORD lpReserved,
    '  LPDWORD lpType,
    '  LPBYTE lpData,
    '  LPDWORD lpcbData
    ');
    Declare Auto Function RegEnumValue Lib "Advapi32" ( _
        ByVal hKey As IntPtr, _
        ByVal dwIndex As Integer, _
        ByVal lpValueName As StringBuilder, _
        ByRef lpcValueName As Integer, _
        ByVal lpReserved As IntPtr, _
        ByVal lpType As IntPtr, _
        ByVal lpData As IntPtr, _
        ByVal lpcbData As IntPtr _
    ) As Integer

    'LONG RegConnectRegistry(
    '  LPCTSTR lpMachineName,
    '  HKEY hKey,
    '  PHKEY phkResult
    ');
    Declare Auto Function RegConnectRegistry Lib "Advapi32" ( _
        ByVal lpMachineName As String, _
        ByVal hKey As IntPtr, _
        ByRef phkResult As IntPtr _
    ) As Integer

    'LONG RegLoadKey(
    '  HKEY hKey,
    '  LPCTSTR lpSubKey,
    '  LPCTSTR lpFile
    ');
    Private Declare Auto Function RegLoadKey Lib "advapi32.dll" ( _
        ByVal hKey As IntPtr, _
        ByVal lpSubKey As String, _
        ByVal lpFile As String _
    ) As Integer

    'LONG RegUnLoadKey(
    '  HKEY hKey,
    '  LPCTSTR lpSubKey
    ');
    Private Declare Auto Function RegUnLoadKey Lib "advapi32.dll" ( _
        ByVal hKey As IntPtr, _
        ByVal lpSubKey As String _
    ) As Integer

#End Region

    Private hKey As IntPtr
    Private RegKeyName As String
    Private IsRootHive As Boolean
    Private Permissions As Integer

    '
    ' Constructors
    '
    Public Sub New()
        hKey = IntPtr.Zero
        IsRootHive = False
    End Sub

    Public Sub New(ByVal Hive As RegistryHive)
        hKey = New IntPtr(Hive)
        RegKeyName = Hive.ToString
        IsRootHive = True
    End Sub

    '
    ' Properties
    '
    Public ReadOnly Property Name() As String
        Get
            Return RegKeyName
        End Get
    End Property

    ' Future: SubKeyCount()

    Public ReadOnly Property ValueCount() As Integer
        Get
            Return GetValueNames.Length
        End Get
    End Property

    '
    ' Methods
    '
    Public Sub Close()
        If Not IsRootHive Then
            RegCloseKey(hKey)
            hKey = IntPtr.Zero
        End If
    End Sub

    ' Future: CreateSubKey()
    ' Future: DeleteSubKey()
    ' Future: DeleleSubKeyTree()
    ' Future: DeleteValue()
    ' Future: Flush()
    ' Future: GetAccessControl()

    Public Function GetSubKeyNames() As String()
        Dim i, ret, NameSize As Integer
        Dim sc As New StringCollection
        Dim sb As New StringBuilder(MAX_REG_KEYNAME_SIZE + 1)
        Dim ans(-1) As String

        ' quick sanity check
        If hKey.Equals(IntPtr.Zero) Then
            Throw New ApplicationException("Cannot access a closed registry key")
        End If

        Do
            NameSize = MAX_REG_KEYNAME_SIZE + 1
            ret = RegEnumKeyEx(hKey, i, sb, NameSize, IntPtr.Zero, IntPtr.Zero, IntPtr.Zero, IntPtr.Zero)
            If ret <> 0 Then
                Exit Do
            End If
            sc.Add(sb.ToString)
            i += 1
        Loop

        If sc.Count > 0 Then
            ReDim ans(sc.Count - 1)
            sc.CopyTo(ans, 0)
        End If
        Return ans
    End Function

    ' Overloaded below
    Public Function GetValue(ByVal Name As String) As Object
        Return GetValue(Name, Nothing, RegistryValueOptions.None)
    End Function

    ' Overloaded below
    Public Function GetValue(ByVal Name As String, ByVal DefaultValue As Object) As Object
        Return GetValue(Name, DefaultValue, RegistryValueOptions.None)
    End Function

    ' Get a value from a registry key
    Public Function GetValue(ByVal Name As String, ByVal DefaultValue As Object, ByVal Options As RegistryValueOptions) As Object
        Dim ret, RegType, BufSize As Integer
        Dim Buf As IntPtr
        Dim ans As Object = Nothing

        ' quick sanity check
        If hKey.Equals(IntPtr.Zero) Then
            Throw New ApplicationException("Cannot access a closed registry key")
        End If

        ' do a "dry run" to get the type and size requirements
        BufSize = 0
        ret = RegQueryValueEx(hKey, Name, IntPtr.Zero, RegType, IntPtr.Zero, BufSize)
        If ret <> 0 And ret <> ERROR_MORE_DATA Then
            Return DefaultValue
        End If

        ' an "empty" key also returns the default
        If BufSize = 0 Then
            Return DefaultValue
        End If

        ' OK, let's do it for real this time
        Buf = Marshal.AllocHGlobal(BufSize)
        ret = RegQueryValueEx(hKey, Name, IntPtr.Zero, RegType, Buf, BufSize)
        If ret <> 0 Then
            Marshal.FreeHGlobal(Buf)
            Return DefaultValue
        End If

        Select Case RegType
            Case REG_SZ
                Dim temp As String
                temp = Marshal.PtrToStringAuto(Buf)
                ans = temp
            Case REG_EXPAND_SZ
                Dim temp As String
                temp = Marshal.PtrToStringAuto(Buf)
                If Options <> RegistryValueOptions.DoNotExpandEnvironmentNames Then
                    temp = Environment.ExpandEnvironmentVariables(temp)
                End If
                ans = temp
            Case REG_DWORD
                Dim temp As Integer
                temp = Marshal.ReadInt32(Buf)
                ans = temp
            Case REG_QWORD
                Dim temp As Long
                temp = Marshal.ReadInt64(Buf)
                ans = temp
            Case REG_BINARY
                Dim temp(BufSize - 1) As Byte
                For i As Integer = 0 To BufSize - 1
                    temp(i) = Marshal.ReadByte(Buf, i)
                Next
                ans = temp
            Case REG_MULTI_SZ
                Dim temp, tmp(-1) As String
                Dim sc As New StringCollection
                Dim ptr As IntPtr

                ptr = Buf
                Do
                    temp = Marshal.PtrToStringAuto(ptr)
                    If temp.Length = 0 Then
                        Exit Do
                    End If
                    sc.Add(temp)
                    ptr = New IntPtr(ptr.ToInt64 + (temp.Length + 1) * Marshal.SystemDefaultCharSize)
                Loop

                If sc.Count > 0 Then
                    ReDim tmp(sc.Count - 1)
                    sc.CopyTo(tmp, 0)
                End If
                ans = tmp
            Case Else
                Marshal.FreeHGlobal(Buf)
                Throw New ApplicationException("Registry type of " & RegType & " is not supported")
        End Select

        Marshal.FreeHGlobal(Buf)
        Return ans
    End Function

    ' Future: GetValueKind()

    Public Function GetValueNames() As String()
        Dim i, ret, NameSize As Integer
        Dim sc As New StringCollection
        Dim sb As New StringBuilder(MAX_REG_KEYNAME_SIZE + 1)
        Dim ans(-1) As String

        ' quick sanity check
        If hKey.Equals(IntPtr.Zero) Then
            Throw New ApplicationException("Cannot access a closed registry key")
        End If

        Do
            NameSize = MAX_REG_KEYNAME_SIZE + 1
            ret = RegEnumValue(hKey, i, sb, NameSize, IntPtr.Zero, IntPtr.Zero, IntPtr.Zero, IntPtr.Zero)
            If ret <> 0 Then
                Exit Do
            End If
            sc.Add(sb.ToString)
            i += 1
        Loop

        If sc.Count > 0 Then
            ReDim ans(sc.Count - 1)
            sc.CopyTo(ans, 0)
        End If
        Return ans
    End Function

    ' static method
    Public Shared Function OpenRemoteBaseKey(ByVal Hive As RegistryHive, ByVal RemotePC As String) As RegistryKey
        Dim ret As Integer
        Dim hRemoteKey As IntPtr

        ret = RegConnectRegistry("\\" & RemotePC, New IntPtr(Hive), hRemoteKey)
        If ret <> 0 Then
            Throw New Win32Exception(ret)
        End If

        Dim ans As New RegistryKey
        ans.hKey = hRemoteKey
        ans.RegKeyName = Hive.ToString
        ans.IsRootHive = False
        ans.Permissions = 0
        Return ans
    End Function

    ' overloaded, see below
    Public Function OpenSubKey(ByVal Name As String) As RegistryKey
        Return OpenSubKey(Name, False, RegWow64Options.None)
    End Function

    ' overloaded, see below
    Public Function OpenSubKey(ByVal Name As String, ByVal Writeable As Boolean) As RegistryKey
        Return OpenSubKey(Name, Writeable, RegWow64Options.None)
    End Function

    ' open a SubKey
    Public Function OpenSubKey(ByVal Name As String, ByVal Writeable As Boolean, ByVal Options As RegWow64Options) As RegistryKey
        Dim ret, Rights As Integer
        Dim hSubKey As IntPtr

        ' quick sanity check
        If hKey.Equals(IntPtr.Zero) Then
            Throw New ApplicationException("Cannot access a closed registry key")
        End If

        Rights = RegistryRights.ReadKey
        If Writeable Then
            Rights = RegistryRights.ReadKey Or RegistryRights.WriteKey
        End If

        ' Ignore the 64-bit flags when running from <= Win2k
        If CDbl(Environment.OSVersion.Version.Major & "." & Environment.OSVersion.Version.Minor) <= 5.0 Then
            Options = RegWow64Options.None
        End If

        ret = RegOpenKeyEx(hKey, Name, 0, Rights Or Options, hSubKey)
        If ret <> 0 Then
            Return Nothing
        End If

        Dim ans As New RegistryKey
        ans.hKey = hSubKey
        ans.RegKeyName = Name
        ans.IsRootHive = False
        ans.Permissions = Rights Or Options
        Return ans
    End Function

    ' Future: SetAccessControl()

    ' Overloaded, see below
    Public Sub SetValue(ByVal Name As String, ByVal Value As Object)
        SetValue(Name, Value, RegistryValueKind.Unknown)
    End Sub

    ' Write a value to a key
    Public Sub SetValue(ByVal Name As String, ByVal Value As Object, ByVal RegType As RegistryValueKind)
        Dim gch As GCHandle
        Dim ptr As IntPtr
        Dim ret, Size As Integer

        ' check out the permissions
        If (Permissions And RegistryRights.WriteKey) = 0 Then
            Throw New ApplicationException("Cannot write to registry key")
        End If

        ' So we have to figure out the type
        If RegType = RegistryValueKind.Unknown Then
            Select Case Value.GetType.ToString
                Case "System.String"
                    RegType = RegistryValueKind.String
                Case "System.Int32"
                    RegType = RegistryValueKind.DWord
                Case "System.Int64"
                    RegType = RegistryValueKind.QWord
                Case "System.String[]"
                    RegType = RegistryValueKind.MultiString
                Case "System.Byte[]"
                    RegType = RegistryValueKind.Binary
                Case Else
                    ' convert it to a string
                    RegType = RegistryValueKind.String
                    Value = Value.ToString
            End Select
        End If

        Select Case RegType
            Case RegistryValueKind.Binary
                Dim temp() As Byte
                temp = DirectCast(Value, System.Byte())
                Size = temp.Length
                gch = GCHandle.Alloc(temp, GCHandleType.Pinned)
                ptr = Marshal.UnsafeAddrOfPinnedArrayElement(temp, 0)
            Case RegistryValueKind.DWord
                Dim temp As Integer = CInt(Value)
                Size = 4
                ptr = Marshal.AllocHGlobal(Size)
                Marshal.WriteInt32(ptr, 0, temp)
            Case RegistryValueKind.ExpandString
                Dim temp As String = CStr(Value)
                Size = (temp.Length + 1) * Marshal.SystemDefaultCharSize
                ptr = Marshal.StringToHGlobalAuto(temp)
            Case RegistryValueKind.MultiString
                Dim temp, lines() As String
                Dim index As Integer
                lines = DirectCast(Value, System.String())

                ' Calculate the total size, including the terminating null
                Size = 0
                For Each temp In lines
                    Size += (temp.Length + 1) * Marshal.SystemDefaultCharSize
                Next
                Size += Marshal.SystemDefaultCharSize
                ptr = Marshal.AllocHGlobal(Size)

                index = 0
                For Each temp In lines
                    Dim tempPtr As IntPtr
                    Dim tempArray() As Char

                    tempArray = temp.ToCharArray
                    tempPtr = New IntPtr(ptr.ToInt64 + index)
                    Marshal.Copy(tempArray, 0, tempPtr, tempArray.Length)
                    index += (tempArray.Length + 1) * Marshal.SystemDefaultCharSize
                Next
            Case RegistryValueKind.QWord
                Dim temp As Long = CLng(Value)
                Size = 8
                ptr = Marshal.AllocHGlobal(Size)
                Marshal.WriteInt64(ptr, 0, temp)
            Case RegistryValueKind.String
                Dim temp As String = CStr(Value)
                Size = (temp.Length + 1) * Marshal.SystemDefaultCharSize
                ptr = Marshal.StringToHGlobalAuto(temp)
            Case Else
                Throw New ApplicationException("Registry type of " & RegType & " is not supported")
        End Select

        ' let's do it!
        ret = RegSetValueEx(hKey, Name, 0, RegType, ptr, Size)

        ' clean up
        If gch.IsAllocated Then
            gch.Free()
        Else
            Marshal.FreeHGlobal(ptr)
        End If

        If ret <> 0 Then
            Throw New Win32Exception(ret)
        End If
    End Sub

    '
    ' Extra methods not in the original .Net Version
    '
    Public Shared Sub LoadKey(ByVal key As RegistryKey, ByVal MountPoint As String, ByVal HivePath As String)
        Dim ret As Integer
        ret = RegLoadKey(key.hKey, MountPoint, HivePath)
        If ret <> 0 Then
            Throw New Win32Exception(ret)
        End If
    End Sub

    Public Shared Sub UnLoadKey(ByVal Key As RegistryKey, ByVal MountPoint As String)
        Dim ret As Integer
        ret = RegUnLoadKey(Key.hKey, MountPoint)
        If ret <> 0 Then
            Throw New Win32Exception(ret)
        End If
    End Sub
End Class

'
' The static Registry Class
'
Public Class Registry
    Public Shared ReadOnly Property LocalMachine() As RegistryKey
        Get
            Return New RegistryKey(RegistryHive.LocalMachine)
        End Get
    End Property

    Public Shared ReadOnly Property ClassesRoot() As RegistryKey
        Get
            Return New RegistryKey(RegistryHive.ClassesRoot)
        End Get
    End Property

    Public Shared ReadOnly Property Users() As RegistryKey
        Get
            Return New RegistryKey(RegistryHive.Users)
        End Get
    End Property

    Public Shared ReadOnly Property CurrentUser() As RegistryKey
        Get
            Return New RegistryKey(RegistryHive.CurrentUser)
        End Get
    End Property

    Public Shared ReadOnly Property CurrentConfig() As RegistryKey
        Get
            Return New RegistryKey(RegistryHive.CurrentConfig)
        End Get
    End Property

    ' Future: SetValue()
    ' Future: GetValue()
End Class


