' Add a COM-based reference to Task Scheduler 1.1 Type Library (for WinVista)
Imports System.Runtime.InteropServices
Imports System.Text
Imports System.XML
Imports System.IO
Imports System.Globalization

'
' Gather information about scheduled tasks
'

' Use the following Compile-time constant to reduce the number of
' entries in the Task table (since WinVista has quite a few internal
' scheduled tasks)
#Const TRIM_MICROSOFT_TASKS = False

Public Class ScheduledTasks

#Region "API Region"
    'typedef enum _TASK_TRIGGER_TYPE {
    '  TASK_TIME_TRIGGER_ONCE = 0,
    '  TASK_TIME_TRIGGER_DAILY = 1,
    '  TASK_TIME_TRIGGER_WEEKLY = 2,
    '  TASK_TIME_TRIGGER_MONTHLY = 3,
    '  TASK_TIME_TRIGGER_MONTHLYDOW = 4,
    '  TASK_EVENT_TRIGGER_ON_IDLE = 5,
    '  TASK_EVENT_TRIGGER_AT_SYSTEMSTART = 6,
    '  TASK_EVENT_TRIGGER_AT_LOGON = 7
    '} TASK_TRIGGER_TYPE, *PTASK_TRIGGER_TYPE;
    Private Enum TASK_TRIGGER_TYPE As Integer
        TASK_TIME_TRIGGER_ONCE = 0
        TASK_TIME_TRIGGER_DAILY
        TASK_TIME_TRIGGER_WEEKLY
        TASK_TIME_TRIGGER_MONTHLY
        TASK_TIME_TRIGGER_MONTHLYDOW
        TASK_EVENT_TRIGGER_ON_IDLE
        TASK_EVENT_TRIGGER_AT_SYSTEMSTART
        TASK_EVENT_TRIGGER_AT_LOGON
    End Enum

    Private Enum TASK_FLAG As Integer
        TASK_FLAG_INTERACTIVE = &H1
        TASK_FLAG_DELETE_WHEN_DONE = &H2
        TASK_FLAG_DISABLED = &H4
        TASK_FLAG_START_ONLY_IF_IDLE = &H10
        TASK_FLAG_KILL_ON_IDLE_END = &H20
        TASK_FLAG_DONT_START_IF_ON_BATTERIES = &H40
        TASK_FLAG_KILL_IF_GOING_ON_BATTERIES = &H80
        TASK_FLAG_RUN_ONLY_IF_DOCKED = &H100
        TASK_FLAG_HIDDEN = &H200
        TASK_FLAG_RUN_IF_CONNECTED_TO_INTERNET = &H400
        TASK_FLAG_RESTART_ON_IDLE_RESUME = &H800
        TASK_FLAG_SYSTEM_REQUIRED = &H1000
        TASK_FLAG_RUN_ONLY_IF_LOGGED_ON = &H2000
    End Enum

    'typedef struct _DAILY {
    '  WORD DaysInterval;
    '} DAILY;
    Private Structure DAILY
        Dim DaysInterval As Short
    End Structure

    'typedef struct _WEEKLY {
    '  WORD WeeksInterval;
    '  WORD rgfDaysOfTheWeek;
    '} WEEKLY;
    Private Structure WEEKLY
        Dim WeeksInterval As Short
        Dim rgfDaysOfTheWeek As Short
    End Structure

    'typedef struct _MONTHLYDATE {
    '  DWORD rgfDays;
    '  WORD rgfMonths;
    '} MONTHLYDATE;
    Private Structure MONTHLYDATE
        Dim rgfDays As Integer
        Dim rgfMonths As Short
    End Structure

    'typedef struct _MONTHLYDOW {
    '  WORD wWhichWeek;
    '  WORD rgfDaysOfTheWeek;
    '  WORD rgfMonths;
    '} MONTHLYDOW;
    Private Structure MONTHLYDOW
        Dim wWhichWeek As Short
        Dim rgfDaysOfTheWeek As Short
        Dim rgfMonths As Short
    End Structure

    'typedef union _TRIGGER_TYPE_UNION {
    '  DAILY Daily;
    '  WEEKLY Weekly;
    '  MONTHLYDATE MonthlyDate;
    '  MONTHLYDOW MonthlyDOW;
    '} TRIGGER_TYPE_UNION;
    <StructLayout(LayoutKind.Explicit)> _
    Private Structure TRIGGER_TYPE_UNION
        <FieldOffset(0)> _
        Dim Daily As DAILY
        <FieldOffset(0)> _
        Dim Weekly As WEEKLY
        <FieldOffset(0)> _
        Dim MonthlyDate As MONTHLYDATE
        <FieldOffset(0)> _
        Dim MonthlyDOW As MONTHLYDOW
    End Structure

    'typedef struct _TASK_TRIGGER {
    '  WORD cbTriggerSize;
    '  WORD Reserved1;
    '  WORD wBeginYear;
    '  WORD wBeginMonth;
    '  WORD wBeginDay;
    '  WORD wEndYear;
    '  WORD wEndMonth;
    '  WORD wEndDay;
    '  WORD wStartHour;
    '  WORD wStartMinute;
    '  DWORD MinutesDuration;
    '  DWORD MinutesInterval;
    '  DWORD rgFlags;
    '  TASK_TRIGGER_TYPE TriggerType;
    '  TRIGGER_TYPE_UNION Type;
    '  WORD Reserved2;
    '  WORD wRandomMinutesInterval;
    '} TASK_TRIGGER, *PTASK_TRIGGER;
    Private Structure TASK_TRIGGER
        Dim cbTriggerSize As Short
        Dim Reserved1 As Short
        Dim wBeginYear As Short
        Dim wBeginMonth As Short
        Dim wBeginDay As Short
        Dim wEndYear As Short
        Dim wEndMonth As Short
        Dim wEndDay As Short
        Dim wStartHour As Short
        Dim wStartMinute As Short
        Dim MinutesDuration As Integer
        Dim MinutesInterval As Integer
        Dim rgFlags As Integer
        Dim TriggerType As TASK_TRIGGER_TYPE
        Dim Type As TRIGGER_TYPE_UNION
        Dim Reserved2 As Short
        Dim wRandomMinutesInterval As Short
    End Structure

    'typedef struct _SYSTEMTIME {
    '  WORD wYear;
    '  WORD wMonth;
    '  WORD wDayOfWeek;
    '  WORD wDay;
    '  WORD wHour;
    '  WORD wMinute;
    '  WORD wSecond;
    '  WORD wMilliseconds;
    '} SYSTEMTIME, *PSYSTEMTIME;
    Private Structure SYSTEMTIME
        Dim wYear As Short
        Dim wMonth As Short
        Dim wDayOfWeek As Short
        Dim wDay As Short
        Dim wHour As Short
        Dim wMinute As Short
        Dim wSecond As Short
        Dim wMilliseconds As Short
    End Structure

    <ComImport(), Guid("148BD52B-A2AB-11CE-B11F-00AA00530503"), InterfaceType(ComInterfaceType.InterfaceIsIUnknown)> _
    Private Interface ITaskTrigger
        'HRESULT SetTrigger(
        '  const PTASK_TRIGGER pTrigger
        ');
        Function SetTrigger(ByRef pTrigger As TASK_TRIGGER) As Integer
        'HRESULT GetTrigger(
        '  PTASK_TRIGGER pTrigger
        ');
        Function GetTrigger(ByRef pTrigger As TASK_TRIGGER) As Integer
        'HRESULT GetTriggerString(
        '  LPWSTR* ppwszTrigger
        ');
        Function GetTriggerString(ByRef ppwszTrigger As IntPtr) As Integer
    End Interface

    <ComImport(), Guid("148BD528-A2AB-11CE-B11F-00AA00530503"), InterfaceType(ComInterfaceType.InterfaceIsIUnknown)> _
    Private Interface IEnumWorkItems
        'HRESULT Next(
        '  ULONG celt,
        '  LPWSTR** rgpwszNames,
        '  ULONG* pceltFetched
        ');
        Function [Next](ByVal celt As Integer, ByRef rgpwszNames As IntPtr, ByRef pceltFetched As Integer) As Integer
        'HRESULT Skip(
        '  ULONG celt
        ');
        Function Skip(ByVal celt As Integer) As Integer
        'HRESULT Reset();
        Function Reset() As Integer
        'HRESULT Clone(
        '  IEnumWorkItems** ppEnumWorkItems
        ');
        Function Clone(ByRef ppEnumWorkItems As IntPtr) As Integer
    End Interface

    <ComImport(), Guid("a6b952f0-a4b1-11d0-997d-00aa006887ec"), InterfaceType(ComInterfaceType.InterfaceIsIUnknown)> _
    Private Interface IScheduledWorkItem
        'HRESULT CreateTrigger(
        '  WORD* piNewTrigger,
        '  ITaskTrigger** ppTrigger
        ');
        Function CreateTrigger(ByRef piNewTrigger As Short, ByRef ppTrigger As IntPtr) As Integer
        'HRESULT DeleteTrigger(
        '  WORD iTrigger
        ');
        Function DeleteTrigger(ByVal iTrigger As Short) As Integer
        'HRESULT GetTriggerCount(
        '  WORD* plCount
        ');
        Function GetTriggerCount(ByRef plCount As Short) As Integer
        'HRESULT GetTrigger(
        '  WORD iTrigger,
        '  ITaskTrigger** ppTrigger
        ');
        Function GetTrigger(ByVal iTrigger As Short, ByRef ppTrigger As IntPtr) As Integer
        'HRESULT GetTriggerString(
        '  WORD iTrigger,
        '  LPWSTR* ppwszTrigger
        ');
        Function GetTriggerString(ByVal iTrigger As Short, <MarshalAs(UnmanagedType.LPWStr)> ByRef ppwszTrigger As String) As Integer
        'HRESULT GetRunTimes(
        '  const LPSYSTEMTIME pstBegin,
        '  const LPSYSTEMTIME pstEnd,
        '  WORD* pCount,
        '  LPSYSTEMTIME* rgstTaskTimes
        ');
        Function GetRunTimes(ByVal pstBegin As SYSTEMTIME, ByVal pstEnd As SYSTEMTIME, ByRef pCount As Short, ByVal rgstTaskTimes As IntPtr) As Integer
        'HRESULT GetNextRunTime(
        '  SYSTEMTIME* pstNextRun
        ');
        Function GetNextRunTime(ByVal pstNextRun As SYSTEMTIME) As Integer
        'HRESULT SetIdleWait(
        '  WORD wIdleMinutes,
        '  WORD wDeadlineMinutes
        ');
        Function SetIdleWait(ByVal wIdleMinutes As Short, ByVal wDeadlineMinutes As Short) As Integer
        'HRESULT GetIdleWait(
        '  WORD* pwIdleMinutes,
        '  WORD* pwDeadlineMinutes
        ');
        Function GetIdleWait(ByRef pwIdleMinutes As Short, ByRef pwDeadlineMinutes As Short) As Integer
        'HRESULT Run();
        Function Run() As Integer
        'HRESULT Terminate();
        Function Terminate() As Integer
        'HRESULT EditWorkItem(
        '  HWND hParent,
        '  DWORD dwReserved
        ');
        Function EditWorkItem(ByVal hParent As IntPtr, ByVal dwReserved As Integer) As Integer
        'HRESULT GetMostRecentRunTime(
        '  SYSTEMTIME* pstLastRun
        ');
        Function GetMostRecentRunTime(ByRef pstLastRun As SYSTEMTIME) As Integer
        'HRESULT GetStatus(
        '  HRESULT* phrStatus
        ');
        Function GetStatus(ByRef phrStatus As Integer) As Integer
        'HRESULT GetExitCode(
        '  DWORD* pdwExitCode
        ');
        Function GetExitCode(ByRef pdwExitCode As Integer) As Integer
        'HRESULT SetComment(
        '  LPCWSTR pwszComment
        ');
        Function SetComment(<MarshalAs(UnmanagedType.LPWStr)> ByVal pwszComment As String) As Integer
        'HRESULT GetComment(
        '  LPWSTR* ppwszComment
        ');
        Function GetComment(<MarshalAs(UnmanagedType.LPWStr)> ByRef ppwszComment As String) As Integer
        'HRESULT SetCreator(
        '  LPCWSTR pwszCreator
        ');
        Function SetCreator(<MarshalAs(UnmanagedType.LPWStr)> ByVal pwszCreator As String) As Integer
        'HRESULT GetCreator(
        '  LPWSTR* ppwszCreator
        ');
        Function GetCreator(<MarshalAs(UnmanagedType.LPWStr)> ByRef ppwszCreator As String) As Integer
        'HRESULT SetWorkItemData(
        '  WORD cBytes,
        '  BYTE rgbData[]
        ');
        Function SetWorkItemData(ByVal cBytes As Short, ByVal rgbData As IntPtr) As Integer
        'HRESULT GetWorkItemData(
        '  WORD* pcBytes,
        '  BYTE** ppBytes
        ');
        Function GetWorkItemData(ByRef pcBytes As Short, ByRef ppBytes As IntPtr) As Integer
        'HRESULT SetErrorRetryCount(
        '  WORD wRetryCount
        ');
        Function SetErrorRetryCount(ByVal wRetryCount As Short) As Integer
        'HRESULT GetErrorRetryCount(
        '  WORD* pwRetryCount
        ');
        Function GetErrorRetryCount(ByRef pwRetryCount As Short) As Integer
        'HRESULT SetErrorRetryInterval(
        '  WORD wRetryInterval
        ');
        Function SetErrorRetryInterval(ByVal wRetryInterval As Short) As Integer
        'HRESULT GetErrorRetryInterval(
        '  WORD* pwRetryInterval
        ');
        Function GetErrorRetryInterval(ByRef pwRetryInterval As Short) As Integer
        'HRESULT SetFlags(
        '  DWORD dwFlags
        ');
        Function SetFlags(ByVal dwFlags As Integer) As Integer
        'HRESULT GetFlags(
        '  DWORD* pdwFlags
        ');
        Function GetFlags(ByRef pdwFlags As Integer) As Integer
        'HRESULT SetAccountInformation(
        '  LPCWSTR pwszAccountName,
        '  LPCWSTR pwszPassword
        ');
        Function SetAccountInformation(<MarshalAs(UnmanagedType.LPWStr)> ByVal pwszAccountName As String, <MarshalAs(UnmanagedType.LPWStr)> ByVal pwszPassword As String) As Integer
        'HRESULT GetAccountInformation(
        '  LPWSTR* ppwszAccountName
        ');
        Function GetAccountInformation(<MarshalAs(UnmanagedType.LPWStr)> ByRef ppwszAccountName As String) As Integer
    End Interface

    <ComImport(), Guid("148BD524-A2AB-11CE-B11F-00AA00530503"), InterfaceType(ComInterfaceType.InterfaceIsIUnknown)> _
    Private Interface ITask
        'HRESULT CreateTrigger(
        '  WORD* piNewTrigger,
        '  ITaskTrigger** ppTrigger
        ');
        Function CreateTrigger(ByRef piNewTrigger As Short, ByRef ppTrigger As IntPtr) As Integer
        'HRESULT DeleteTrigger(
        '  WORD iTrigger
        ');
        Function DeleteTrigger(ByVal iTrigger As Short) As Integer
        'HRESULT GetTriggerCount(
        '  WORD* plCount
        ');
        Function GetTriggerCount(ByRef plCount As Short) As Integer
        'HRESULT GetTrigger(
        '  WORD iTrigger,
        '  ITaskTrigger** ppTrigger
        ');
        Function GetTrigger(ByVal iTrigger As Short, ByRef ppTrigger As IntPtr) As Integer
        'HRESULT GetTriggerString(
        '  WORD iTrigger,
        '  LPWSTR* ppwszTrigger
        ');
        Function GetTriggerString(ByVal iTrigger As Short, <MarshalAs(UnmanagedType.LPWStr)> ByRef ppwszTrigger As String) As Integer
        'HRESULT GetRunTimes(
        '  const LPSYSTEMTIME pstBegin,
        '  const LPSYSTEMTIME pstEnd,
        '  WORD* pCount,
        '  LPSYSTEMTIME* rgstTaskTimes
        ');
        Function GetRunTimes(ByVal pstBegin As SYSTEMTIME, ByVal pstEnd As SYSTEMTIME, ByRef pCount As Short, ByVal rgstTaskTimes As IntPtr) As Integer
        'HRESULT GetNextRunTime(
        '  SYSTEMTIME* pstNextRun
        ');
        Function GetNextRunTime(ByVal pstNextRun As SYSTEMTIME) As Integer
        'HRESULT SetIdleWait(
        '  WORD wIdleMinutes,
        '  WORD wDeadlineMinutes
        ');
        Function SetIdleWait(ByVal wIdleMinutes As Short, ByVal wDeadlineMinutes As Short) As Integer
        'HRESULT GetIdleWait(
        '  WORD* pwIdleMinutes,
        '  WORD* pwDeadlineMinutes
        ');
        Function GetIdleWait(ByRef pwIdleMinutes As Short, ByRef pwDeadlineMinutes As Short) As Integer
        'HRESULT Run();
        Function Run() As Integer
        'HRESULT Terminate();
        Function Terminate() As Integer
        'HRESULT EditWorkItem(
        '  HWND hParent,
        '  DWORD dwReserved
        ');
        Function EditWorkItem(ByVal hParent As IntPtr, ByVal dwReserved As Integer) As Integer
        'HRESULT GetMostRecentRunTime(
        '  SYSTEMTIME* pstLastRun
        ');
        Function GetMostRecentRunTime(ByRef pstLastRun As SYSTEMTIME) As Integer
        'HRESULT GetStatus(
        '  HRESULT* phrStatus
        ');
        Function GetStatus(ByRef phrStatus As Integer) As Integer
        'HRESULT GetExitCode(
        '  DWORD* pdwExitCode
        ');
        Function GetExitCode(ByRef pdwExitCode As Integer) As Integer
        'HRESULT SetComment(
        '  LPCWSTR pwszComment
        ');
        Function SetComment(<MarshalAs(UnmanagedType.LPWStr)> ByVal pwszComment As String) As Integer
        'HRESULT GetComment(
        '  LPWSTR* ppwszComment
        ');
        Function GetComment(<MarshalAs(UnmanagedType.LPWStr)> ByRef ppwszComment As String) As Integer
        'HRESULT SetCreator(
        '  LPCWSTR pwszCreator
        ');
        Function SetCreator(<MarshalAs(UnmanagedType.LPWStr)> ByVal pwszCreator As String) As Integer
        'HRESULT GetCreator(
        '  LPWSTR* ppwszCreator
        ');
        Function GetCreator(<MarshalAs(UnmanagedType.LPWStr)> ByRef ppwszCreator As String) As Integer
        'HRESULT SetWorkItemData(
        '  WORD cBytes,
        '  BYTE rgbData[]
        ');
        Function SetWorkItemData(ByVal cBytes As Short, ByVal rgbData As IntPtr) As Integer
        'HRESULT GetWorkItemData(
        '  WORD* pcBytes,
        '  BYTE** ppBytes
        ');
        Function GetWorkItemData(ByRef pcBytes As Short, ByRef ppBytes As IntPtr) As Integer
        'HRESULT SetErrorRetryCount(
        '  WORD wRetryCount
        ');
        Function SetErrorRetryCount(ByVal wRetryCount As Short) As Integer
        'HRESULT GetErrorRetryCount(
        '  WORD* pwRetryCount
        ');
        Function GetErrorRetryCount(ByRef pwRetryCount As Short) As Integer
        'HRESULT SetErrorRetryInterval(
        '  WORD wRetryInterval
        ');
        Function SetErrorRetryInterval(ByVal wRetryInterval As Short) As Integer
        'HRESULT GetErrorRetryInterval(
        '  WORD* pwRetryInterval
        ');
        Function GetErrorRetryInterval(ByRef pwRetryInterval As Short) As Integer
        'HRESULT SetFlags(
        '  DWORD dwFlags
        ');
        Function SetFlags(ByVal dwFlags As Integer) As Integer
        'HRESULT GetFlags(
        '  DWORD* pdwFlags
        ');
        Function GetFlags(ByRef pdwFlags As Integer) As Integer
        'HRESULT SetAccountInformation(
        '  LPCWSTR pwszAccountName,
        '  LPCWSTR pwszPassword
        ');
        Function SetAccountInformation(<MarshalAs(UnmanagedType.LPWStr)> ByVal pwszAccountName As String, <MarshalAs(UnmanagedType.LPWStr)> ByVal pwszPassword As String) As Integer
        'HRESULT GetAccountInformation(
        '  LPWSTR* ppwszAccountName
        ');
        Function GetAccountInformation(<MarshalAs(UnmanagedType.LPWStr)> ByRef ppwszAccountName As String) As Integer
        'HRESULT SetApplicationName(
        '  LPCWSTR pwszApplicationName
        ');
        Function SetApplicationName(<MarshalAs(UnmanagedType.LPWStr)> ByVal pwszApplicationName As String) As Integer
        'HRESULT GetApplicationName(
        '  LPWSTR* ppwszApplicationName
        ');
        Function GetApplicationName(<MarshalAs(UnmanagedType.LPWStr)> ByRef ppwszApplicationName As String) As Integer
        'HRESULT SetParameters(
        '  LPCWSTR pwszParameters
        ');
        Function SetParameter(<MarshalAs(UnmanagedType.LPWStr)> ByVal pwszParameters As String) As Integer
        'HRESULT GetParameters(
        '  LPWSTR* ppwszParameters
        ');
        Function GetParameters(<MarshalAs(UnmanagedType.LPWStr)> ByRef ppwszParameters As String) As Integer
        'HRESULT SetWorkingDirectory(
        '  LPCWSTR pwszWorkingDirectory
        ');
        Function SetWorkingDirectory(<MarshalAs(UnmanagedType.LPWStr)> ByVal pwszWorkingDirectory As String) As Integer
        'HRESULT GetWorkingDirectory(
        '  LPWSTR* ppwszWorkingDirectory
        ');
        Function GetWorkingDirectory(ByVal ppwszWorkingDirectory As String) As Integer
        'HRESULT SetPriority(
        '  DWORD dwPriority
        ');
        Function SetPriority(ByVal dwPriority As Integer) As Integer
        'HRESULT GetPriority(
        '  DWORD* pdwPriority
        ');
        Function GetPriority(ByRef pdwPriority As Integer) As Integer
        'HRESULT SetTaskFlags(
        '  DWORD dwFlags
        ');
        Function SetTaskFlags(ByVal dwFlags As Integer) As Integer
        'HRESULT GetTaskFlags(
        '  DWORD* pdwFlags
        ');
        Function GetTaskFlags(ByRef pdwFlags As Integer) As Integer
        'HRESULT SetMaxRunTime(
        '  DWORD dwMaxRunTime
        ');
        Function SetMaxRunTime(ByVal dwMaxRunTime As Integer) As Integer
        'HRESULT GetMaxRunTime(
        '  DWORD* pdwMaxRunTime
        ');
        Function GetMaxRunTime(ByRef pdwMaxRunTime As Integer) As Integer
    End Interface

    <ComImport(), Guid("148BD527-A2AB-11CE-B11F-00AA00530503"), InterfaceType(ComInterfaceType.InterfaceIsIUnknown)> _
    Private Interface ITaskScheduler
        'HRESULT SetTargetComputer(
        '  LPCWSTR pwszComputer
        ');
        Function SetTargetComputer(<MarshalAs(UnmanagedType.LPWStr)> ByVal pwszComputer As String) As Integer
        'HRESULT GetTargetComputer(
        '  LPWSTR* ppwszComputer
        ');
        Function GetTargetComputer(<MarshalAs(UnmanagedType.LPWStr)> ByRef ppwszComputer As String) As Integer
        'HRESULT Enum(
        '  IEnumWorkItems** ppEnumTasks
        ');
        Function [Enum](<MarshalAs(UnmanagedType.Interface)> ByRef ppEnumTasks As IEnumWorkItems) As Integer
        'HRESULT Activate(
        '  LPCWSTR pwszName,
        '  REFIID riid,
        '  IUnknown** ppunk
        ');
        Function Activate(<MarshalAs(UnmanagedType.LPWStr)> ByVal pwszName As String, ByRef riid As System.Guid, <MarshalAs(UnmanagedType.IUnknown)> ByRef ppunk As Object) As Integer
        'HRESULT Delete(
        '  LPCWSTR pwszName
        ');
        Function Delete(<MarshalAs(UnmanagedType.LPWStr)> ByVal pwszName As String) As Integer
        'HRESULT NewWorkItem(
        '  LPCWSTR pwszTaskName,
        '  REFCLSID rclsid,
        '  REFIID riid,
        '  IUnknown** ppunk
        ');
        Function NewWorkItem(<MarshalAs(UnmanagedType.LPWStr)> ByVal pwszTaskName As String, ByVal rclsid As IntPtr, ByVal riid As System.Guid, <MarshalAs(UnmanagedType.IUnknown)> ByRef ppunk As IntPtr) As Integer
        'HRESULT AddWorkItem(
        '  LPCWSTR pwszTaskName,
        '  IScheduledWorkItem* pWorkItem
        ');
        Function AddWorkItem(<MarshalAs(UnmanagedType.LPWStr)> ByVal pwszTaskName As String, <MarshalAs(UnmanagedType.Interface)> ByRef pWorkItem As IScheduledWorkItem) As Integer
        'HRESULT IsOfType(
        '  LPCWSTR pwszName,
        '  REFIID riid
        ');
        Function IfOfType(<MarshalAs(UnmanagedType.LPWStr)> ByVal pwszName As String, ByVal riid As System.Guid) As Integer
    End Interface

    Private Const SCHED_S_TASK_READY As Integer = &H41300
    Private Const SCHED_S_TASK_RUNNING As Integer = &H41301
    Private Const SCHED_S_TASK_DISABLED As Integer = &H41302
    Private Const SCHED_S_TASK_HAS_NOT_RUN As Integer = &H41303
    Private Const SCHED_S_TASK_NOT_SCHEDULED As Integer = &H41305

#End Region

    Private Const BUFSIZE As Integer = 4096

    '
    ' Use of bunch of low-level API routines to get the status of the
    ' scheduled tasks.  Writes directly to the datatable provided.
    '
    Public Sub GetTaskInfo(ByVal MachineName As String, ByVal ID As String, ByVal WinDir As String, ByVal dt As SOSOSDataSet.TasksDataTable)
        Dim CLSID_CTaskScheduler, ITaskGUID, IScheduledWorkItemGUID As Guid
        Dim CTaskSchedulerType As Type
        Dim iSched As ITaskScheduler
        Dim iTask As ITask
        Dim p As IntPtr
        Dim Temp, Buf, TaskName, UNC As String
        Dim iWork As IEnumWorkItems
        Dim fetched As Integer
        Dim obj As Object
        Dim i, cnt As Integer
        Dim s, tc As Short
        Dim sb As StringBuilder
        Dim st As SYSTEMTIME
        Dim culture As CultureInfo
        Dim dr As SOSOSDataSet.TasksRow

        culture = New CultureInfo("en-US", True)

        ITaskGUID = New Guid("{148BD524-A2AB-11CE-B11F-00AA00530503}")
        IScheduledWorkItemGUID = New Guid("{a6b952f0-a4b1-11d0-997d-00aa006887ec}")

        CLSID_CTaskScheduler = New Guid("{148BD52A-A2AB-11CE-B11F-00AA00530503}")
        CTaskSchedulerType = Type.GetTypeFromCLSID(CLSID_CTaskScheduler, True)
        iSched = CType(Activator.CreateInstance(CTaskSchedulerType), ITaskScheduler)

        ' Get the location/name of the Task directory
        UNC = Misc.ConvUNC(MachineName, WinDir & "\tasks")

        ' connect to the PC (use this form even on the local PC)
        iSched.SetTargetComputer("\\" & MachineName)

        iWork = Nothing
        iSched.Enum(iWork)

        ' loop thru the list of tasks
        cnt = 0
        obj = Nothing
        Do While Not CBool(iWork.Next(1, p, fetched))
            If fetched = 0 Then
                Exit Do
            End If

            ' Get the TaskName
            TaskName = Marshal.PtrToStringUni(Marshal.ReadIntPtr(p))
            Marshal.FreeCoTaskMem(Marshal.ReadIntPtr(p))
            Marshal.FreeCoTaskMem(p)

            ' Get iTask object
            iSched.Activate(TaskName, ITaskGUID, obj)
            iTask = DirectCast(obj, ITask)

            dr = dt.NewTasksRow
            dr.ID_Tasks = ID
            dr.Task_ID = cnt
            dr.Task_JobName = Left(TaskName.Replace(".job", ""), dt.Task_JobNameColumn.MaxLength)

            iTask.GetStatus(i)
            dr.Task_State = "Scheduled"
            dr.Task_Status = ""
            Select Case i
                Case SCHED_S_TASK_DISABLED
                    dr.Task_State = "Disabled"
                Case SCHED_S_TASK_HAS_NOT_RUN
                    dr.Task_Status = "Never Run"
                Case SCHED_S_TASK_RUNNING
                    ' status (not state!) field
                    dr.Task_Status = "Running"
            End Select

            ' The API interface doesn't support getting the "extended status", so we
            ' have to poke around the file to get that
            dr.Task_Status = Left(GetExtStatus(UNC & "\" & TaskName), dt.Task_StatusColumn.MaxLength)

            Buf = ""
            iTask.GetComment(Buf)
            dr.Task_Comment = Left(Buf, dt.Task_CommentColumn.MaxLength)

            Buf = ""
            iTask.GetApplicationName(Buf)
            Temp = Buf

            Buf = ""
            iTask.GetParameters(Buf)
            If Buf.Length > 0 Then
                Temp &= " " & Buf.Trim
            End If
            dr.Task_Command = Left(Temp, dt.Task_CommandColumn.MaxLength)

            iTask.GetTriggerCount(tc)
            sb = New StringBuilder
            For s = 0 To CShort(tc - 1)
                Buf = ""
                iTask.GetTriggerString(s, Buf)
                sb.Append(Buf)
                If s <> tc - 1 Then
                    sb.Append("; ")
                End If
            Next
            dr.Task_Schedule = Left(sb.ToString, dt.Task_ScheduleColumn.MaxLength)

            iTask.GetMostRecentRunTime(st)
            If st.wYear > 0 Then
                ' using the en-US format for the date/time
                dr.Task_LastRun = Date.Parse([String].Format("{0:00}/{1:00}/{2} {3:00}:{4:00}", st.wMonth, st.wDay, st.wYear, st.wHour, st.wMinute), culture)
            End If

            Try
                ' This can throw an exception (but not fail!?!) if the account
                ' information is not correct
                iTask.GetExitCode(i)
            Catch
            End Try
            dr.Task_LastResult = i

            Buf = ""
            Try
                iTask.GetAccountInformation(Buf)
                If Buf = "" Then
                    dr.Task_Account = "SYSTEM"
                Else
                    dr.Task_Account = Left(Buf, dt.Task_AccountColumn.MaxLength)
                End If
            Catch
            End Try

            dt.AddTasksRow(dr)
            cnt += 1
        Loop
    End Sub

    '
    ' Get the "Extended Status".  We have to physically read the task file
    ' to get this information.
    ' Note: that's my name for this field, since it's not documented in the
    ' API, I don't really know what it's real name is.
    '
    Private Shared Function GetExtStatus(ByVal JobFile As String) As String
        Dim buf(BUFSIZE) As Byte
        Dim fs As FileStream
        Dim buf_size, block_num, block_start, block_size As Integer
        Dim StatusCode As Integer

        Try
            ' read the file into a buffer
            fs = File.OpenRead(JobFile)
            buf_size = fs.Read(buf, 0, BUFSIZE)
            fs.Close()

            ' "walk" the first 4 blocks
            block_num = 0
            block_start = &H46
            Do While block_start < buf_size
                block_size = BitConverter.ToInt16(buf, block_start)
                block_start = block_start + (block_size * 2) + 2
                block_num += 1
                If block_num > 4 Then
                    Exit Do
                End If
            Loop

            ' the Detailed Status code (at the 5th block)
            StatusCode = BitConverter.ToInt32(buf, block_start + 4)
            If StatusCode <> 0 And StatusCode <> SCHED_S_TASK_HAS_NOT_RUN Then
                Dim ex As New System.ComponentModel.Win32Exception(StatusCode)
                Return ex.Message
            End If
        Catch
        End Try

        Return ""
    End Function

    '
    ' Get Scheduled Task information from WinVista
    '
    Sub GetTaskInfoVista(ByVal MachineName As String, ByVal ID As String, ByVal dt As SOSOSDataSet.TasksDataTable)
#If BUILD_ON_WINDOWS_VISTA Then
        Dim ts As New TaskScheduler.TaskScheduler
        Dim folder As TaskScheduler.ITaskFolder

        If MachineName <> "" Then
            ts.Connect(MachineName)
        Else
            ts.Connect()
        End If

        folder = ts.GetFolder("\")
        DoFolder(folder, ID, dt, 0)
#End If
    End Sub

#If BUILD_ON_WINDOWS_VISTA Then
    '
    ' A recursive routine to get task info from entries in the folder
    '
    Sub DoFolder(ByVal obj As Object, ByVal ID As String, ByVal dt As SOSOSDataSet.TasksDataTable, ByRef cnt As Integer)
        Dim dr As SOSOSDataSet.TasksRow
        Dim rtc As TaskScheduler.IRegisteredTaskCollection
        Dim rt As TaskScheduler.IRegisteredTask
        Dim trigger As TaskScheduler.ITrigger
        Dim action As TaskScheduler.IAction
        Dim SubFolder As TaskScheduler.ITaskFolder
        Dim Schedule, Command As String
        Dim folder As TaskScheduler.ITaskFolder

        folder = DirectCast(obj, TaskScheduler.ITaskFolder)
        rtc = folder.GetTasks(0)
        For Each rt In rtc
            dr = dt.NewTasksRow
            dr.ID_Tasks = ID
            dr.Task_ID = cnt
            dr.Task_JobName = Left(rt.Path, dt.Task_JobNameColumn.MaxLength)
            dr.Task_Comment = Left(rt.Definition.RegistrationInfo.Description, dt.Task_CommentColumn.MaxLength)

            Select Case rt.State
                Case TaskScheduler._TASK_STATE.TASK_STATE_DISABLED
                    dr.Task_State = "Disabled"
                Case TaskScheduler._TASK_STATE.TASK_STATE_QUEUED
                    dr.Task_State = "Scheduled"
                Case TaskScheduler._TASK_STATE.TASK_STATE_READY
                    dr.Task_State = "Scheduled"
                Case TaskScheduler._TASK_STATE.TASK_STATE_RUNNING
                    dr.Task_State = "Scheduled"
                    dr.Task_Status = "Running"
                Case TaskScheduler._TASK_STATE.TASK_STATE_UNKNOWN
                    dr.Task_State = "Unknown"
            End Select

            If rt.LastRunTime.Year > 1970 Then
                dr.Task_LastRun = rt.LastRunTime
                dr.Task_LastResult = rt.LastTaskResult
                If rt.LastTaskResult <> 0 Then
                    Dim ex As New System.ComponentModel.Win32Exception(rt.LastTaskResult)
                    dr.Task_Status = Left(ex.Message, dt.Task_StatusColumn.MaxLength)
                End If
            Else
                dr.Task_Status = "Never Run"
            End If

            ' credentials for this task
            If rt.Definition.Principal.Id = "Author" Then
                dr.Task_Account = Left(rt.Definition.Principal.UserId, dt.Task_AccountColumn.MaxLength)
            Else
                dr.Task_Account = Left(rt.Definition.Principal.Id, dt.Task_AccountColumn.MaxLength)
            End If

            ' there could be multiple schedules (triggers) per task, and
            ' some tasks are "manual" and have no trigger
            Schedule = ""
            For Each trigger In rt.Definition.Triggers
                Select Case trigger.Type
                    Case TaskScheduler._TASK_TRIGGER_TYPE2.TASK_TRIGGER_BOOT
                        Schedule &= "Boot; "
                    Case TaskScheduler._TASK_TRIGGER_TYPE2.TASK_TRIGGER_LOGON
                        Schedule &= "Login; "
                    Case TaskScheduler._TASK_TRIGGER_TYPE2.TASK_TRIGGER_EVENT
                        Schedule &= "Event; "
                    Case TaskScheduler._TASK_TRIGGER_TYPE2.TASK_TRIGGER_IDLE
                        Schedule &= "Idle; "
                    Case TaskScheduler._TASK_TRIGGER_TYPE2.TASK_TRIGGER_DAILY
                        Dim daily As TaskScheduler.IDailyTrigger
                        daily = CType(trigger, TaskScheduler.IDailyTrigger)
                        Schedule &= "At " & CDate(daily.StartBoundary).ToShortTimeString & " Daily, starting " & CDate(daily.StartBoundary).ToShortDateString
                        If daily.DaysInterval > 1 Then
                            Schedule &= " every " & daily.DaysInterval & " days"
                        End If
                        Schedule &= "; "
                    Case TaskScheduler._TASK_TRIGGER_TYPE2.TASK_TRIGGER_MONTHLY
                        Dim monthly As TaskScheduler.IMonthlyTrigger
                        Dim i, mask As Integer

                        monthly = CType(trigger, TaskScheduler.IMonthlyTrigger)
                        Schedule &= "At " & CDate(monthly.StartBoundary).ToShortTimeString
                        If monthly.DaysOfMonth >= 2147483647 Then
                            Schedule &= " every day"
                        Else
                            Schedule &= " on day "
                            For i = 1 To 31
                                mask = CInt(2 ^ (i - 1))
                                If (mask And monthly.DaysOfMonth) > 0 Then
                                    Schedule &= i & ","
                                End If
                            Next
                        End If
                        Schedule = Schedule.TrimEnd(","c)
                        Schedule &= " of "
                        If monthly.MonthsOfYear = 4095 Then
                            Schedule &= "every month"
                        Else
                            For i = 1 To 12
                                mask = CInt(2 ^ (i - 1))
                                If (mask And monthly.MonthsOfYear) > 0 Then
                                    Schedule &= MonthName(i, True) & ","
                                End If
                            Next
                        End If
                        Schedule = Schedule.TrimEnd(","c)
                        Schedule &= ", starting " & CDate(trigger.StartBoundary).ToShortDateString & "; "
                    Case TaskScheduler._TASK_TRIGGER_TYPE2.TASK_TRIGGER_MONTHLYDOW
                        Dim monthdoy As TaskScheduler.IMonthlyDOWTrigger
                        Dim i, mask As Integer

                        monthdoy = CType(trigger, TaskScheduler.IMonthlyDOWTrigger)
                        Schedule &= "At " & CDate(monthdoy.StartBoundary).ToShortTimeString
                        If monthdoy.WeeksOfMonth >= 15 Then
                            Schedule &= " every"
                        Else
                            For i = 1 To 5
                                mask = CInt(2 ^ (i - 1))
                                If (mask And monthdoy.WeeksOfMonth) > 0 Then
                                    Select Case i
                                        Case 1
                                            Schedule &= " First,"
                                        Case 2
                                            Schedule &= " Second,"
                                        Case 3
                                            Schedule &= " Third,"
                                        Case 4
                                            Schedule &= " Forth,"
                                        Case 5
                                            Schedule &= " Last,"
                                    End Select
                                End If
                            Next
                        End If
                        Schedule = Schedule.TrimEnd(","c)
                        If monthdoy.DaysOfWeek >= 127 Then
                            Schedule &= " weekday"
                        Else
                            For i = 1 To 12
                                mask = CInt(2 ^ (i - 1))
                                If (mask And monthdoy.DaysOfWeek) > 0 Then
                                    Schedule &= " " & WeekdayName(i, True) & ","
                                End If
                            Next
                        End If
                        Schedule = Schedule.TrimEnd(","c)
                        Schedule &= " of "
                        If monthdoy.MonthsOfYear >= 4095 Then
                            Schedule &= "every month"
                        Else
                            For i = 1 To 12
                                mask = CInt(2 ^ (i - 1))
                                If (mask And monthdoy.MonthsOfYear) > 0 Then
                                    Schedule &= MonthName(i, True) & ","
                                End If
                            Next
                        End If
                        Schedule = Schedule.TrimEnd(","c)
                        Schedule &= ", starting " & CDate(trigger.StartBoundary).ToShortDateString & "; "
                    Case TaskScheduler._TASK_TRIGGER_TYPE2.TASK_TRIGGER_WEEKLY
                        Dim weekly As TaskScheduler.IWeeklyTrigger
                        Dim i, mask As Integer
                        weekly = CType(trigger, TaskScheduler.IWeeklyTrigger)
                        Schedule &= "At " & CDate(weekly.StartBoundary).ToShortTimeString & " every "
                        If weekly.DaysOfWeek >= 127 Then
                            Schedule &= "weekday"
                        Else
                            For i = 1 To 12
                                mask = CInt(2 ^ (i - 1))
                                If (mask And weekly.DaysOfWeek) > 0 Then
                                    Schedule &= WeekdayName(i, True) & ","
                                End If
                            Next
                        End If
                        Schedule = Schedule.TrimEnd(","c)
                        Schedule &= ", starting " & CDate(weekly.StartBoundary).ToShortDateString & "; "
                    Case TaskScheduler._TASK_TRIGGER_TYPE2.TASK_TRIGGER_TIME
                        Schedule &= "At " & CDate(trigger.StartBoundary).ToShortTimeString & " on " & CDate(trigger.StartBoundary).ToShortDateString & "; "
                    Case TaskScheduler._TASK_TRIGGER_TYPE2.TASK_TRIGGER_REGISTRATION
                        Schedule &= "Registration; "
                    Case TaskScheduler._TASK_TRIGGER_TYPE2.TASK_TRIGGER_SESSION_STATE_CHANGE
                        Schedule &= "Session; "
                End Select
            Next
            If Schedule <> "" Then
                dr.Task_Schedule = Left(Schedule.Trim.TrimEnd(";"c), dt.Task_ScheduleColumn.MaxLength)
            End If

            ' the Action types
            Command = ""
            For Each action In rt.Definition.Actions
                Select Case action.Type
                    Case TaskScheduler._TASK_ACTION_TYPE.TASK_ACTION_EXEC
                        ' Oddly enough, the actual command itself is not available
                        ' via the TaskScheduler interface. So, we have to parse the
                        ' XML.
                        Dim xml_doc As Xml.XmlDocument
                        Dim xml_nsm As XmlNamespaceManager
                        Dim xml_node As Xml.XmlNode
                        Dim temp As String

                        xml_doc = New Xml.XmlDocument
                        xml_doc.LoadXml(rt.Xml)
                        xml_nsm = New XmlNamespaceManager(xml_doc.NameTable)
                        xml_nsm.AddNamespace("tsk", "http://schemas.microsoft.com/windows/2004/02/mit/task")

                        xml_node = xml_doc.SelectSingleNode("//tsk:Exec/tsk:Command", xml_nsm)
                        temp = ""
                        If Not IsNothing(xml_node) Then
                            temp = xml_node.InnerXml
                        End If

                        xml_node = xml_doc.SelectSingleNode("//tsk:Exec/tsk:Arguments", xml_nsm)
                        If Not IsNothing(xml_node) Then
                            temp &= " " & xml_node.InnerXml
                        End If
                        Command &= temp & "; "
                    Case TaskScheduler._TASK_ACTION_TYPE.TASK_ACTION_SEND_EMAIL
                        Command &= "Email; "
                    Case TaskScheduler._TASK_ACTION_TYPE.TASK_ACTION_SHOW_MESSAGE
                        Command &= "Message; "
                    Case TaskScheduler._TASK_ACTION_TYPE.TASK_ACTION_COM_HANDLER
                        Command &= "Handler; "
                End Select
            Next
            If Command <> "" Then
                dr.Task_Command = Left(Command.Trim.TrimEnd(";"c), dt.Task_CommandColumn.MaxLength)
            End If

            dt.AddTasksRow(dr)
            cnt += 1
        Next

        ' a recursive call to other folders
        For Each SubFolder In folder.GetFolders(0)
            ' let's skip the Microsoft folder
#If TRIM_MICROSOFT_TASKS Then
            If Not SubFolder.Path.StartsWith("\Microsoft") Then
#End If
            DoFolder(SubFolder, ID, dt, cnt)
#If TRIM_MICROSOFT_TASKS Then
            End If
#End If
        Next
    End Sub
#End If
End Class
