' Add a COM reference to Microsoft ActiveX Data Objects Library (ADO)
' Add a COM reference to Microsoft ADO Ext. for DDL and Security (ADOX)
Imports System.IO

'
' Copy a detached dataSet (of possibly "foreign" origin) to a Microsoft Access
' database file.  If the database doesn't exit, it will be created.  If the
' table doesn't exist, it will be created.
'
Class dt2mdb

    Public Sub CopyToMDB(ByVal dt As DataTable, ByVal dbName As String)
        CopyToMDB(dt, dbName, False, "")
    End Sub

    Public Sub CopyToMDB(ByVal dt As DataTable, ByVal dbName As String, ByVal EmptyRowsFirst As Boolean)
        CopyToMDB(dt, dbName, EmptyRowsFirst, "")
    End Sub

    Public Sub CopyToMDB(ByVal dt As DataTable, ByVal dbName As String, ByVal EmptyRowsFirst As Boolean, ByVal tblName As String)
        Dim con As New ADODB.Connection
        Dim cat As New ADOX.Catalog
        Dim rs As New ADODB.Recordset
        Dim tbl As ADOX.Table
        Dim col As ADOX.Column
        Dim key As ADOX.Key
        Dim dr As DataRow
        Dim dc As DataColumn
        Dim got_it As Boolean
        Dim size, i As Integer

        ' If the mdb file already exits, then that's OK by me.  Otherwise create
        ' the mdb file using ADOX
        If Not File.Exists(dbName) Then
            cat.Create("Provider=Microsoft.Jet.OLEDB.4.0;Data Source=" & dbName)
            con = CType(cat.ActiveConnection, ADODB.Connection)
        End If

        ' Open a connection if it isn't already open
        If con.State = ADODB.ObjectStateEnum.adStateClosed Then
            con.Open("Provider=Microsoft.Jet.OLEDB.4.0;Data Source=" & dbName)
            cat.ActiveConnection = con
        End If

        ' Check to see if the Table exists, use the default name if not specified
        If tblName = "" Then
            If IsNothing(dt.TableName) OrElse dt.TableName = "" Then
                tblName = "Table1"
            Else
                tblName = dt.TableName
            End If
        End If
        got_it = False
        For Each tbl In cat.Tables
            If tbl.Name.ToLower = tblName.ToLower Then
                got_it = True
                Exit For
            End If
        Next

        ' Create the new table if required.  Try to match the native Access data types as
        ' best as possible.
        If Not got_it Then
            tbl = New ADOX.Table
            tbl.Name = tblName
            tbl.ParentCatalog = cat
            For Each dc In dt.Columns
                If dc.MaxLength <= 0 Then
                    size = 255
                Else
                    size = dc.MaxLength
                End If
                col = New ADOX.Column
                col.Name = dc.ColumnName
                Select Case dc.DataType.ToString
                    Case "System.Boolean"
                        col.Type = ADOX.DataTypeEnum.adBoolean
                        tbl.Columns.Append(col)
                    Case "System.Byte"
                        col.Type = ADOX.DataTypeEnum.adUnsignedTinyInt
                        tbl.Columns.Append(col)
                        tbl.Columns(col.Name).Properties("Default").Value = 0
                    Case "System.Byte[]"
                        col.Type = ADOX.DataTypeEnum.adLongVarBinary
                        tbl.Columns.Append(col)
                    Case "System.Char"
                        If size > 255 Then
                            ' use memo field if over 255
                            col.Type = ADOX.DataTypeEnum.adLongVarWChar
                        Else
                            col.Type = ADOX.DataTypeEnum.adVarWChar
                            col.DefinedSize = size
                        End If
                        tbl.Columns.Append(col)
                        tbl.Columns(col.Name).Properties("Jet OLEDB:Compressed UniCode Strings").Value = True
                    Case "System.DateTime"
                        col.Type = ADOX.DataTypeEnum.adDate
                        tbl.Columns.Append(col)
                    Case "System.Decimal"
                        col.Type = ADOX.DataTypeEnum.adNumeric
                        tbl.Columns.Append(col)
                        tbl.Columns(col.Name).Precision = 18
                        tbl.Columns(col.Name).NumericScale = 4
                        tbl.Columns(col.Name).Properties("Default").Value = 0
                    Case "System.Double"
                        col.Type = ADOX.DataTypeEnum.adDouble
                        tbl.Columns.Append(col)
                        tbl.Columns(col.Name).Properties("Default").Value = 0
                    Case "System.Int16"
                        col.Type = ADOX.DataTypeEnum.adSmallInt
                        tbl.Columns.Append(col)
                        tbl.Columns(col.Name).Properties("Default").Value = 0
                    Case "System.Int32"
                        col.Type = ADOX.DataTypeEnum.adInteger
                        tbl.Columns.Append(col)
                        tbl.Columns(col.Name).Properties("Default").Value = 0
                    Case "System.Int64"
                        ' obvious problems with possible overflow
                        col.Type = ADOX.DataTypeEnum.adInteger
                        tbl.Columns.Append(col)
                        tbl.Columns(col.Name).Properties("Default").Value = 0
                    Case "System.SByte"
                        col.Type = ADOX.DataTypeEnum.adUnsignedTinyInt
                        tbl.Columns.Append(col)
                        tbl.Columns(col.Name).Properties("Default").Value = 0
                    Case "System.Single"
                        col.Type = ADOX.DataTypeEnum.adSingle
                        tbl.Columns.Append(col)
                        tbl.Columns(col.Name).Properties("Default").Value = 0
                    Case "System.String"
                        If size > 255 Then
                            ' use memo field if over 255
                            col.Type = ADOX.DataTypeEnum.adLongVarWChar
                        Else
                            col.Type = ADOX.DataTypeEnum.adVarWChar
                            col.DefinedSize = size
                        End If
                        tbl.Columns.Append(col)
                        tbl.Columns(col.Name).Properties("Jet OLEDB:Compressed UniCode Strings").Value = True
                    Case "System.TimeSpan"
                        col.Type = ADOX.DataTypeEnum.adDate
                        tbl.Columns.Append(col)
                    Case "System.UInt16"
                        ' no unsigned types in Access, oh well...
                        col.Type = ADOX.DataTypeEnum.adSmallInt
                        tbl.Columns.Append(col)
                        tbl.Columns(col.Name).Properties("Default").Value = 0
                    Case "System.UInt32"
                        col.Type = ADOX.DataTypeEnum.adInteger
                        tbl.Columns.Append(col)
                        tbl.Columns(col.Name).Properties("Default").Value = 0
                    Case "System.Unt64"
                        col.Type = ADOX.DataTypeEnum.adInteger
                        tbl.Columns.Append(col)
                        tbl.Columns(col.Name).Properties("Default").Value = 0
                    Case Else
                        col.Type = ADOX.DataTypeEnum.adLongVarBinary
                        tbl.Columns.Append(col)
                End Select

                If dc.DataType Is GetType(System.Boolean) = False Then
                    If dc.AllowDBNull Then
                        tbl.Columns(col.Name).Properties("Nullable").Value = True
                    Else
                        tbl.Columns(col.Name).Properties("Nullable").Value = False
                    End If
                End If

                If dc.AutoIncrement Then
                    tbl.Columns(col.Name).Properties("Autoincrement").Value = True
                    tbl.Columns(col.Name).Properties("Seed").Value = CInt(dc.AutoIncrementSeed)
                    tbl.Columns(col.Name).Properties("Increment").Value = CInt(dc.AutoIncrementStep)
                End If
            Next

            ' recreate the primary key
            If UBound(dt.PrimaryKey) >= 0 Then
                key = New ADOX.Key
                key.Name = "PK_" & tblName
                key.Type = ADOX.KeyTypeEnum.adKeyPrimary
                For Each dc In dt.PrimaryKey
                    key.Columns.Append(dc.ColumnName)
                    tbl.Columns(dc.ColumnName).Properties("Nullable").Value = False
                Next
                tbl.Keys.Append(key)
            End If

            cat.Tables.Append(tbl)
        End If

        ' Open an ADODB recordset to the table
        rs.Open(tblName, con, ADODB.CursorTypeEnum.adOpenKeyset, ADODB.LockTypeEnum.adLockOptimistic)

        ' Delete any existing data, if requested
        If EmptyRowsFirst Then
            con.Execute("delete from " & tblName, , ADODB.CommandTypeEnum.adCmdText + ADODB.ExecuteOptionEnum.adExecuteNoRecords)
        End If

        ' OK, now we're finally ready to copy the data!
        For Each dr In dt.Rows
            Try
                rs.AddNew()
                For i = 0 To dt.Columns.Count - 1
                    If Not IsDBNull(dr(i)) Then
                        ' do the best we can with overflow conditions
                        Select Case rs(i).Type.ToString
                            Case "adSmallInt"
                                Try
                                    rs(i).Value = CShort(dr(i))
                                Catch
                                    rs(i).Value = 0
                                End Try
                            Case "adInteger"
                                Try
                                    rs(i).Value = CInt(dr(i))
                                Catch
                                    rs(i).Value = 0
                                End Try
                            Case "adVarWChar"
                                rs(i).Value = Left(dr(i).ToString, rs(i).DefinedSize)
                            Case Else
                                rs(i).Value = dr(i)
                        End Select
                    End If
                Next
                rs.Update()
            Catch ex As Exception
                ' The tables can be dynamically changing during the execution
                ' of this program.  This sometime leads to errors due to the
                ' way we read the table in "chunks".  This is an acceptible
                ' risk considering the reduction in memory footprint.
            End Try
        Next
        rs.Close()
        con.Close()

        con = Nothing
        cat = Nothing
        rs = Nothing
    End Sub
End Class
