Imports System.IO
Imports System.Net
Imports System.Runtime.InteropServices

'
' Poll by list of IP addresses
' Note: Inherited from the fmPoll_Base form
'
Public Class fmPoll_byIP

    Public Overrides Function GetPCList() As String()
        Return IPList(tbInclude.Text, tbExclude.Text)
    End Function

    Private Function IPList(ByVal IP_Include As String, ByVal IP_Exclude As String) As String()
        Dim ip, ip1, ip2 As Long
        Dim line, parts() As String
        Dim cnt, ln, i, j As Integer
        Dim exclude As Boolean
        Dim IPinclude(10), IPexclude(10), ans(10) As String
        Dim sr As StringReader

        ' parse the include ranges...
        sr = New StringReader(IP_Include)
        ln = 1
        cnt = 0
        line = sr.ReadLine
        While Not line Is Nothing
            parts = Split(line, "-")
            If UBound(parts) = 1 Then
                ip1 = IPStrToLong(parts(0).Trim)
                ip2 = IPStrToLong(parts(1).Trim)
                If ip1 = 0 Or ip2 = 0 Then
                    MsgBox("Error parsing IP include address at line " & ln, MsgBoxStyle.Exclamation, "Warning")
                    Return Nothing
                End If
                If ip2 < ip1 Then
                    MsgBox("IP include address out of range at line " & ln, MsgBoxStyle.Exclamation, "Warning")
                    Return Nothing
                End If
                ' add these to our list
                For ip = ip1 To ip2
                    IPinclude(cnt) = IPLongToStr(ip)
                    cnt += 1
                    ' keep from thrashing the redim... we increment by 10
                    If cnt > UBound(IPinclude) Then
                        ReDim Preserve IPinclude(cnt + 9)
                    End If
                Next
            End If
            If UBound(parts) = 0 Then
                ip1 = IPStrToLong(parts(0).Trim)
                If ip1 = 0 Then
                    MsgBox("Error parsing IP include address at line " & ln, MsgBoxStyle.Exclamation, "Warning")
                    Return Nothing
                End If
                IPinclude(cnt) = IPLongToStr(ip1)
                cnt += 1
                ' keep from thrashing the redim... we increment by 10
                If cnt > UBound(IPinclude) Then
                    ReDim Preserve IPinclude(cnt + 9)
                End If
            End If
            ln += 1
            line = sr.ReadLine
        End While
        ReDim Preserve IPinclude(cnt - 1)

        ' now parse the exclude ranges...
        sr = New StringReader(IP_Exclude)
        ln = 1
        cnt = 0
        line = sr.ReadLine
        While Not line Is Nothing
            parts = Split(line, "-")
            If UBound(parts) = 1 Then
                ip1 = IPStrToLong(parts(0).Trim)
                ip2 = IPStrToLong(parts(1).Trim)
                If ip1 = 0 Or ip2 = 0 Then
                    MsgBox("Error parsing IP exclude address at line " & ln, MsgBoxStyle.Exclamation, "Warning")
                    Return Nothing
                End If
                If ip2 < ip1 Then
                    MsgBox("IP exclude address out of range at line " & ln, MsgBoxStyle.Exclamation, "Warning")
                    Return Nothing
                End If
                ' add these to our list
                For ip = ip1 To ip2
                    IPexclude(cnt) = IPLongToStr(ip)
                    cnt += 1
                    ' keep from thrashing the redim... we increment by 10
                    If cnt > UBound(IPexclude) Then
                        ReDim Preserve IPexclude(cnt + 9)
                    End If
                Next
            End If
            If UBound(parts) = 0 Then
                ip1 = IPStrToLong(parts(0).Trim)
                If ip1 = 0 Then
                    MsgBox("Error parsing IP exclude address at line " & ln, MsgBoxStyle.Exclamation, "Warning")
                    Return Nothing
                End If
                IPexclude(cnt) = IPLongToStr(ip1)
                cnt += 1
                ' keep from thrashing the redim... we increment by 10
                If cnt > UBound(IPexclude) Then
                    ReDim Preserve IPexclude(cnt + 9)
                End If
            End If
            ln += 1
            line = sr.ReadLine
        End While
        ReDim Preserve IPexclude(cnt - 1)

        ' merge the two lists together...
        cnt = 0
        For i = 0 To UBound(IPinclude)
            exclude = False
            For j = 0 To UBound(IPexclude)
                If IPinclude(i) = IPexclude(j) Then
                    exclude = True
                    Exit For
                End If
            Next
            If Not exclude Then
                ans(cnt) = IPinclude(i)
                cnt += 1
                If cnt > UBound(ans) Then
                    ReDim Preserve ans(cnt + 9)
                End If
            End If
        Next

        If cnt = 0 Then
            Return Nothing
        End If

        ReDim Preserve ans(cnt - 1)
        Return ans
    End Function

    ' VB.Net doesn't support a Union type, but it does allow you to 
    ' "decorate" a structure to accomplish the same thing
    <StructLayout(LayoutKind.Explicit)> _
    Private Structure IP_Parts
        <FieldOffset(0)> Dim address As Int64
        <FieldOffset(3)> Dim Right As Byte
        <FieldOffset(2)> Dim middleRight As Byte
        <FieldOffset(1)> Dim middleLeft As Byte
        <FieldOffset(0)> Dim Left As Byte
    End Structure

    Private Shared Function IPStrToLong(ByVal IPstr As String) As Long
        Dim ip As IPAddress
        Dim parts As IP_Parts
        Dim buf() As String

        ip = [IPAddress].None
        If Not [IPAddress].TryParse(IPstr, ip) Then
            Return 0
        End If

        ' The IPAddress.Address property has been "depreciated", but we still
        ' need a way to convert a IPv4 "quad dot" to and from a long, so just
        ' keep the compiler happy, we use this routine

        buf = ip.ToString.Split("."c)
        If buf.Length <> 4 Then
            ' Sorry, we don't support IPv6 notation
            Return 0
        End If

        parts.Right = CByte(buf(3))
        parts.middleRight = CByte(buf(2))
        parts.middleLeft = CByte(buf(1))
        parts.Left = CByte(buf(0))

        Return parts.address
    End Function

    Private Shared Function IPLongToStr(ByVal IPlong As Long) As String
        Dim parts As IP_Parts
        Dim buf As String

        parts.address = IPlong
        buf = parts.Left & "." & parts.middleLeft & "." & parts.middleRight & "." & parts.Right

        Return buf
    End Function
End Class
